from .types import (
    t_notebook,
    t_output,
    guess_type,
    copy_button,
    t_add,
    t_update,
    t_delete,
)
from IPython.core.display import display, HTML, JSON, Image, SVG, Markdown
from .manager import Manager
import pandas as pd
import markdown2
import requests
import warnings
import os


class Api:
    naas = None
    role = t_notebook
    manager = None
    deprecated_name = False

    def __init__(self, deprecated_name=False):
        self.deprecated_name = deprecated_name
        self.manager = Manager(t_notebook)
        self.path = self.manager.path

    def deprecatedPrint(self):
        # TODO remove this in june 2021
        if self.deprecated_name:
            warnings.warn(
                "[Warning], naas.api is deprecated,\n use naas.webhook instead it will be remove in 1 june 2021"
            )

    def list(self, path=None):
        self.deprecatedPrint()
        return self.manager.list_prod("list_history", path)

    def list_output(self, path=None):
        self.deprecatedPrint()
        return self.manager.list_prod("list_output", path)

    def get(self, path=None, histo=None):
        self.deprecatedPrint()
        return self.manager.get_file(path, histo=histo)

    def get_output(self, path=None, histo=None):
        self.deprecatedPrint()
        return self.manager.get_file(path, t_output, histo)

    def clear(self, path=None, histo=None):
        self.deprecatedPrint()
        return self.manager.clear_file(path, None, histo)

    def clear_output(self, path=None, histo=None):
        self.deprecatedPrint()
        return self.manager.clear_file(path, t_output, histo)

    def currents(self, raw=False):
        self.deprecatedPrint()
        json_data = self.manager.get_naas()
        if raw:
            json_filtered = []
            for item in json_data:
                if item["type"] == self.role and item["status"] != t_delete:
                    print(item)
                    json_filtered.append(item)
                return json_filtered
        else:
            for item in json_data:
                kind = None
                if item["type"] == self.role and item["status"] != t_delete:
                    kind = f"callable with this url {self.manager.proxy_url('notebooks', item['value'])}"
                    print(f'File ==> {item["path"]} is {kind}')

    def run(self, path=None, debug=False):
        self.deprecatedPrint()
        current_file = self.manager.get_path(path)
        if current_file is None:
            print("Missing file path")
            return
        token = os.urandom(30).hex()
        status = t_add
        try:
            token = self.manager.get_value(current_file, False)
            status = t_update
        except:  # noqa: E722
            pass
        url = self.manager.proxy_url(self.role, token)
        if self.manager.is_production():
            print("No add done, you are in production\n")
            return url
        self.manager.add_prod(
            {
                "type": self.role,
                "status": status,
                "path": current_file,
                "params": {},
                "value": token,
            },
            debug,
        )
        try:
            r = requests.get(url)
            r.raise_for_status()
            print("👌 Well done! Your Notebook runned in production.\n")
        except Exception:
            print("😢 Your Notebook failed in production.\n")
        self.get_output(path)
        self.manager.del_prod({"type": self.role, "path": current_file}, debug)
        return url

    def find(self, path=None):
        current_file = self.manager.get_path(path)
        if current_file is None:
            print("Missing file path")
            return
        try:
            token = self.manager.get_value(current_file, False)
            return self.manager.proxy_url(self.role, token)
        except:  # noqa: E722
            return None

    def add(self, path=None, params={}, debug=False):
        self.deprecatedPrint()
        current_file = self.manager.get_path(path)
        if current_file is None:
            print("Missing file path")
            return
        token = os.urandom(30).hex()
        status = t_add
        try:
            token = self.manager.get_value(current_file, False)
            status = t_update
        except:  # noqa: E722
            pass
        url = self.manager.proxy_url(self.role, token)
        if self.manager.is_production():
            print("No add done, you are in production\n")
            return url
        self.manager.add_prod(
            {
                "type": self.role,
                "status": status,
                "path": current_file,
                "params": params,
                "value": token,
            },
            debug,
        )
        print("👌 Well done! Your Notebook has been sent to production.\n")
        copy_button(url)
        print(
            'PS: to remove the "Notebook as API" feature, just replace .add by .delete'
        )
        return url

    def respond_notebook(self):
        self.deprecatedPrint()
        display(
            Markdown(
                "Response Set as Notebook !",
                metadata={"naas_api": True, "naas_type": t_notebook},
            )
        )

    def respond_file(self, path):
        self.deprecatedPrint()
        abs_path = os.path.abspath(path)
        naas_type = guess_type(abs_path)
        display(Markdown("Response Set as File, preview below: "))
        display(
            JSON(
                {"path": abs_path}, metadata={"naas_api": True, "naas_type": naas_type}
            )
        )

    def respond_html(self, data):
        self.deprecatedPrint()
        display(Markdown("Response Set as HTML, preview below: "))
        display(HTML(data, metadata={"naas_api": True}))

    def respond_json(self, data):
        self.deprecatedPrint()
        display(Markdown("Response Set as JSON, preview below: "))
        display(JSON(data, metadata={"naas_api": True}))

    def respond_image(self, data=None, filename=None):
        self.deprecatedPrint()
        display(Markdown("Response Set as IMAGE, preview below: "))
        display(Image(data, filename=filename, metadata={"naas_api": True}))

    def respond_svg(self, data):
        self.deprecatedPrint()
        display(Markdown("Response Set as SVG, preview below: "))
        display(SVG(data, metadata={"naas_api": True}))

    def respond_text(self, data):
        self.deprecatedPrint()
        display(Markdown("Response Set as Text, preview below: "))
        display(HTML(data, metadata={"naas_api": True, "naas_type": "text"}))

    def respond_markdown(self, data):
        self.deprecatedPrint()
        display(Markdown("Response Set as Markdown, preview below: "))
        html = markdown2.markdown(data)
        display(HTML(html, metadata={"naas_api": True, "naas_type": "markdown"}))

    def respond_csv(self, data):
        self.deprecatedPrint()
        if not isinstance(data, pd.DataFrame):
            raise TypeError("data shoud be a dataframe")
        display(HTML(data.to_html(), metadata={"naas_api": True, "naas_type": "csv"}))

    def delete(self, path=None, all=False, debug=False):
        self.deprecatedPrint()
        if self.manager.is_production():
            print("No delete done, you are in production\n")
            return
        current_file = self.manager.get_path(path)
        self.manager.del_prod({"type": self.role, "path": current_file}, debug)
        print("🗑 Done! Your Notebook has been remove from production.\n")
        if all is True:
            self.clear(path)
            self.clear_output(path)



from naas.runner.env_var import n_env
from naas.types import copy_button
import pandas as pd
import requests
import time
import json


class Callback:
    logger = None

    headers = None

    def __init__(self, logger=None):
        #         n_env.callback_api = "http://naas-callback:3004"
        self.headers = {"Authorization": f"token {n_env.token}"}
        self.logger = logger

    def add(self, response=None, responseHeaders=None, autoDelete=True):
        try:
            data = {
                "response": response,
                "autoDelete": autoDelete,
                "responseHeaders": responseHeaders,
            }
            req = None
            req = requests.post(
                url=f"{n_env.callback_api}/", headers=self.headers, json=data
            )
            req.raise_for_status()
            jsn = req.json()
            print("👌 🔙 Callback has been created successfully !")
            url = f"{n_env.callback_api}/{jsn.get('uuid')}"
            copy_button(url)
            return {"url": url, "uuid": jsn.get("uuid")}
        except Exception as err:
            if self.logger is not None:
                self.logger.error(
                    json.dumps({"id": None, "type": "email error", "error": str(err)})
                )
            else:
                print(err)

    def __get(self, uuid):
        try:
            data = {
                "uuid": uuid,
            }
            req = None
            req = requests.get(
                url=f"{n_env.callback_api}/",
                params={"uuid": uuid},
                headers=self.headers,
                json=data,
            )
            req.raise_for_status()
            jsn = req.json()
            return jsn
        except Exception as err:
            if self.logger is not None:
                self.logger.error(
                    json.dumps({"id": uuid, "type": "email error", "error": str(err)})
                )
            else:
                print(err)

    def get(self, uuid, wait_until_data=False, timeout=3000, raw=False):
        data = None
        total = 0
        while data is None or data.get("result") is None:
            if total > timeout:
                print("🥲 Callback Get timeout !")
                return None
            data = self.__get(uuid)
            time.sleep(1)
            total += 1
            if wait_until_data:
                break
        if data and data.get("result") and data.get("result") != "":
            print("👌 🔙 Callback has been trigger, here your data !")
        else:
            print("🥲 Callback is empty !")
        return data if raw else data.get("result")

    def delete(self, uuid):
        try:
            data = {
                "uuid": uuid,
            }
            req = None
            req = requests.delete(
                url=f"{n_env.callback_api}/", headers=self.headers, json=data
            )
            req.raise_for_status()
            print("👌 🔙 Callback has been delete successfully !")
            return
        except Exception as err:
            if self.logger is not None:
                self.logger.error(
                    json.dumps({"id": uuid, "type": "email error", "error": str(err)})
                )
            else:
                print(err)

    def status(self):
        req = requests.get(url=f"{n_env.callback_api}/")
        req.raise_for_status()
        jsn = req.json()
        return jsn

    def list(self):
        req = requests.get(
            url=f"{n_env.callback_api}/",
            headers=self.headers,
        )
        req.raise_for_status()
        jsn = req.json()
        return pd.DataFrame(data=jsn.get("callbacks"))

    def list_all(self):
        req = requests.get(
            url=f"{n_env.callback_api}/admin",
            headers=self.headers,
        )
        req.raise_for_status()
        jsn = req.json()
        return pd.DataFrame(data=jsn.get("callbacks"))



from asyncio import Semaphore
from naas.types import (
    t_delete,
    t_add,
    t_skip,
    t_update,
    t_job,
    t_error,
    t_start,
    t_notebook,
    t_asset,
    t_output,
    t_dependency,
    t_scheduler,
    t_list,
)
from .env_var import n_env
import pandas as pd
import datetime
import errno
import json
import os
import uuid
from sanic.exceptions import ServerError
import pytz

filters = [t_notebook, t_asset, t_dependency, t_scheduler]
filters_api = [t_notebook, t_asset]


class Jobs:
    __storage_sem = None
    __df = None
    __logger = None
    __json_name = "jobs.json"
    __colums = [
        "id",
        "type",
        "value",
        "path",
        "status",
        "params",
        "lastUpdate",
        "lastRun",
        "runs",
    ]

    def __init__(self, logger, clean=False, init_data=[]):
        self.__json_secrets_path = os.path.join(
            n_env.path_naas_folder, self.__json_name
        )
        self.__storage_sem = Semaphore(1)
        self.__logger = logger
        if not os.path.exists(n_env.path_naas_folder):
            try:
                print("Init Naas folder Jobs")
                os.makedirs(n_env.path_naas_folder)
            except OSError as exc:  # Guard against race condition
                print("__path_naas_files", n_env.path_naas_folder)
                if exc.errno != errno.EEXIST:
                    raise
            except Exception as e:
                print("Exception", e)
        if not os.path.exists(self.__json_secrets_path) or clean:
            uid = str(uuid.uuid4())
            try:
                print("Init Job Storage", self.__json_secrets_path)
                self.__save_to_file(uid, init_data)
                self.__df = None
            except Exception as e:
                print("Exception", e)
                self.__logger.error(
                    {
                        "id": uid,
                        "type": "init_job_storage",
                        "filepath": self.__json_secrets_path,
                        "status": "error",
                        "error": str(e),
                    }
                )
        else:
            uid = str(uuid.uuid4())
            self.__df = self.__get_save_from_file(uid)
        self.__cleanup_jobs()

    def reload_jobs(self):
        uid = str(uuid.uuid4())
        self.__df = self.__get_save_from_file(uid)
        self.__cleanup_jobs()

    def __cleanup_jobs(self):
        try:
            if self.__df is None or len(self.__df) == 0:
                self.__df = pd.DataFrame(columns=self.__colums)
            if len(self.__df) > 0:
                self.__df = self.__df[self.__df.type.isin(filters)]
                self.__dedup_jobs()
                self.__migrate_oldPath()
        except Exception as err:
            print("Cannot cleanup", err)

    def __migrate_oldPath(self):
        cur_jobs = self.__df.to_dict("records")
        for job in cur_jobs:
            path = job["path"]
            tmp_path = path.replace(n_env.path_naas_folder, "")
            if not tmp_path.startswith(n_env.server_root) and os.path.exists(path):
                new_path = os.path.join(
                    n_env.path_naas_folder, n_env.server_root, tmp_path
                )
                os.makedirs(os.path.basename(new_path))
                os.rename(path, new_path)
                job["path"] = new_path
            # previous path /home/ftp/.naas/toto/tata.py
            # new path /home/ftp/.naas/home/ftp/toto/tata.py
            # tmp_path = tmp_path.replace(n_env.path_naas_folder, '').replace(f"{n_env.server_root}/", '')
        self.__df = pd.DataFrame(cur_jobs)
        self.__df = self.__df.reset_index(drop=True)

    def __dedup_jobs(self):
        new_df = self.__df[
            (self.__df.type != t_notebook) & (self.__df.type != t_asset)
        ].to_dict("records")
        cur_notebook = self.__df[self.__df.type == t_notebook]
        cur_asset = self.__df[self.__df.type == t_asset]
        cur_asset = cur_asset.drop_duplicates(subset=["value"]).to_dict("records")
        cur_notebook = cur_notebook.drop_duplicates(subset=["value"]).to_dict("records")
        self.__df = pd.DataFrame([*new_df, *cur_asset, *cur_notebook])
        self.__df = self.__df.reset_index(drop=True)

    def __get_save_from_file(self, uid):
        data = []
        try:
            f = open(self.__json_secrets_path, "r")
            data_l = json.load(f)
            f.close()
            # data = data_l
            dt_string = datetime.datetime.now(tz=pytz.timezone(n_env.tz)).strftime(
                "%Y-%m-%d %H:%M:%S"
            )
            for d in data_l:
                # Fix formating of old jobs
                runs = d.get("runs", [])
                if type(runs) != list:
                    try:
                        runs = json.loads(runs)
                    except Exception:
                        runs = []
                        pass
                c = {
                    "id": d.get("id", uid),
                    "type": d.get("type", ""),
                    "value": d.get("value", ""),
                    "path": d.get("path", ""),
                    "status": d.get("status", t_update),
                    "params": d.get("params", {}),
                    "lastUpdate": d.get("lastUpdate", dt_string),
                    "lastRun": d.get("lastRun", 0),
                    "runs": runs,
                }
                data.append(c)
        except Exception as err:
            self.__logger.error(
                {
                    "id": str(uid),
                    "type": "__get_save_from_file",
                    "status": "exception",
                    "filepath": self.__json_secrets_path,
                    "error": str(err),
                }
            )
        return pd.DataFrame(data).reset_index(drop=True)

    def __save_to_file(self, uid, data):
        try:
            with open(self.__json_secrets_path, "w+") as f:
                f.write(
                    json.dumps(data, sort_keys=True, indent=4).replace("NaN", "null")
                )
                f.close()
        except Exception as err:
            print(f"==> Cannot save {uid} \n\n", err)
            self.__logger.error(
                {
                    "id": str(uid),
                    "type": "__save_to_file",
                    "status": "exception",
                    "filepath": self.__json_secrets_path,
                    "error": str(err),
                }
            )

    async def find_by_value(self, uid, value, target_type):
        res = None
        async with self.__storage_sem:
            try:
                if len(self.__df) > 0:
                    cur_jobs = self.__df[
                        (self.__df.type == target_type) & (self.__df.value == value)
                    ]
                    cur_job = cur_jobs.to_dict("records")
                    if len(cur_job) > 0:
                        res = cur_job[0]
            except Exception as e:
                print("find_by_value", e)
            return res

    async def find_by_path(self, uid, filepath, target_type=None):
        res = None
        async with self.__storage_sem:
            try:
                if len(self.__df) > 0:
                    if target_type:
                        cur_jobs = self.__df[
                            (self.__df.type == target_type)
                            & (self.__df.path.str.lower() == filepath.lower())
                        ]
                    else:
                        cur_jobs = self.__df[(self.__df.path == filepath)]
                    cur_job = cur_jobs.to_dict("records")
                    if len(cur_job) > 0:
                        res = cur_job[0]
            except Exception as e:
                print("find_by_path", e)
            return res

    async def is_running(self, uid, notebook_filepath, target_type):
        res = False
        try:
            cur_job = await self.find_by_path(uid, notebook_filepath, target_type)
            if cur_job:
                status = cur_job.get("status", None)
                if status and status == t_start:
                    res = True
        except Exception as e:
            print("is_running", e)
        return res

    def __match_clear(self, cur_filename, filename, clear_all):
        if clear_all and cur_filename.endwith(filename):
            return True
        elif not clear_all and cur_filename == filename:
            return True
        else:
            return False

    def clear_file(self, uid, path, histo, mode=None):
        # possible format
        # histo_filename
        # out_filename
        # histo_out_filename
        filename = os.path.basename(path)
        clear_all = False
        if mode:
            filename = f"{mode}_{filename}"
        if histo and histo == "all":
            clear_all = True
        elif histo:
            filename = f"{histo}_{filename}"
        removed = []
        dirname = os.path.dirname(path)
        if os.path.exists(path):
            for ffile in os.listdir(dirname):
                if self.__match_clear(ffile, filename, clear_all):
                    tmp_path = os.path.join(dirname, ffile)
                    removed.append(tmp_path)
                    self.__logger.info(
                        {
                            "id": uid,
                            "filename": filename,
                            "histo": histo,
                            "type": "clear_file",
                            "status": t_delete,
                            "filepath": path,
                        }
                    )
                    os.remove(tmp_path)
        return removed

    def list_files(self, uid, path, filetype, output=False):
        d = []
        dirname = os.path.dirname(path)
        filename = os.path.basename(path)
        # if output:
        #     filename = f"output_{filetype}_{filename}"
        # else:
        #     filename = f"{filetype}_{filename}"
        if output:
            filename = f"___{t_output}__{filename}"
        else:
            filename = f"___{filename}"
        for ffile in os.listdir(dirname):
            if ffile.endswith(filename):
                split_list = ffile.split("___")
                histo = split_list[0]
                tmp_path = os.path.join(dirname, ffile)
                tmp_path = tmp_path.replace(n_env.path_naas_folder, "").replace(
                    f"{n_env.server_root}/", ""
                )
                d.append({"timestamp": histo, "filepath": tmp_path})
        self.__logger.info(
            {
                "id": uid,
                "type": "clear_file",
                "filename": filename,
                "status": t_list,
                "filepath": path,
            }
        )
        return d

    async def list(self, uid, as_df=False):
        data = []
        try:
            async with self.__storage_sem:
                if as_df:
                    data = self.__df
                else:
                    data = self.__df.to_dict("records")
                    for d in data:
                        try:
                            d["path"] = (
                                d["path"]
                                .replace(n_env.path_naas_folder, "")
                                .replace(f"{n_env.server_root}/", "")
                            )
                            d["runs"] = json.loads(d.get("runs", "[]"))
                        except Exception:
                            d["runs"] = []
                            pass
        except Exception as e:
            print("list", e)
        return data

    def __delete(self, cur_elem, uid, path, target_type, value, params):
        try:
            dt_string = datetime.datetime.now(tz=pytz.timezone(n_env.tz)).strftime(
                "%Y-%m-%d %H:%M:%S"
            )
            self.__logger.info(
                {
                    "id": uid,
                    "type": target_type,
                    "value": value,
                    "status": t_delete,
                    "filepath": path,
                    "params": params,
                }
            )
            index = cur_elem.index[0]
            self.__df.at[index, "id"] = uid
            self.__df.at[index, "status"] = t_delete
            self.__df.at[index, "lastUpdate"] = dt_string
            return t_delete
        except Exception as e:
            self.__logger.error(
                {
                    "id": uid,
                    "type": target_type,
                    "value": value,
                    "status": t_error,
                    "filepath": path,
                    "params": params,
                    "error": str(e),
                }
            )
            return t_error

    def __add(self, uid, path, target_type, value, params, run_time):
        try:
            dt_string = datetime.datetime.now(tz=pytz.timezone(n_env.tz)).strftime(
                "%Y-%m-%d %H:%M:%S"
            )
            self.__logger.info(
                {
                    "id": uid,
                    "type": target_type,
                    "value": value,
                    "status": t_add,
                    "filepath": path,
                    "params": params,
                }
            )
            new_row = {
                "id": uid,
                "type": target_type,
                "value": value,
                "status": t_add,
                "path": path,
                "params": params,
                "lastRun": run_time,
                "runs": json.dumps([]),
                "lastUpdate": dt_string,
            }
            cur_df = self.__df.to_dict("records")
            if len(self.__df) > 0:
                self.__df = pd.DataFrame([*cur_df, new_row])
            else:
                self.__df = pd.DataFrame([new_row])
            return t_add
        except Exception as e:
            self.__logger.error(
                {
                    "id": uid,
                    "type": target_type,
                    "value": value,
                    "status": t_error,
                    "filepath": path,
                    "params": params,
                    "error": str(e),
                }
            )
            return t_error

    def __update(self, cur_elem, uid, path, value, params, status, run_time):
        now = datetime.datetime.now(tz=pytz.timezone(n_env.tz))
        dt_string = now.strftime("%Y-%m-%d %H:%M:%S")
        self.__logger.info(
            {
                "id": uid,
                "type": t_job,
                "value": value,
                "status": t_update,
                "filepath": path,
                "params": params,
            }
        )
        index = cur_elem.index[0]
        self.__df.at[index, "id"] = uid
        self.__df.at[index, "status"] = status
        self.__df.at[index, "value"] = value
        self.__df.at[index, "params"] = params
        self.__df.at[index, "lastUpdate"] = dt_string
        if run_time > 0:
            try:
                runs = json.loads(self.__df.at[index, "runs"])
            except Exception:
                runs = []
                pass
            runs.append(
                {"id": uid, "duration": run_time, "date": dt_string, "status": status}
            )
            self.__df.at[index, "runs"] = json.dumps(runs)
            self.__df.at[index, "lastRun"] = run_time
            return t_update
        else:
            return t_skip

    async def update(self, uid, path, target_type, value, params, status, run_time=0):
        data = None
        res = t_error
        async with self.__storage_sem:
            try:
                cur_elem = self.__df.query(
                    f'type == "{target_type}" and path == "{path}"'
                )
                if len(cur_elem) == 1 and status == t_delete:
                    res = self.__delete(cur_elem, uid, path, target_type, value, params)
                elif len(cur_elem) == 1:
                    res = self.__update(
                        cur_elem,
                        uid,
                        path,
                        value,
                        params,
                        status,
                        run_time,
                    )
                elif len(cur_elem) == 0 and status == t_add:
                    res = self.__add(uid, path, target_type, value, params, run_time)
                else:
                    res = t_skip
                if res == t_error:
                    raise ServerError(
                        {
                            "status": "error",
                            "id": uid,
                            "data": [],
                            "error": "unknow error",
                        },
                        status_code=500,
                    )
                data = self.__df.to_dict("records")
                self.__save_to_file(uid, data)
                return {"id": uid, "status": res, "data": data}
            except Exception as e:
                self.__logger.error(
                    {
                        "id": uid,
                        "type": target_type,
                        "value": value,
                        "status": t_error,
                        "filepath": path,
                        "params": params,
                        "error": str(e),
                    }
                )
                raise ServerError(
                    {"status": "error", "id": uid, "data": [], "error": str(e)},
                    status_code=500,
                )



from sanic.views import HTTPMethodView
from sanic import response
from sanic.exceptions import ServerError
from naas.types import t_asset, t_health, t_error, t_start, t_send, t_delete
import uuid
import os


class AssetsController(HTTPMethodView):
    __logger = None
    __jobs = None
    __path_lib_files = None
    __assets_folder = "assets"

    def __init__(self, logger, jobs, path_assets, *args, **kwargs):
        super(AssetsController, self).__init__(*args, **kwargs)
        self.__logger = logger
        self.__jobs = jobs
        self.__path_lib_files = path_assets

    async def get(self, request, token):
        if token.startswith("naas_"):
            return await response.file(
                os.path.join(self.__path_lib_files, self.__assets_folder, token)
            )
        else:
            uid = str(uuid.uuid4())
            job = await self.__jobs.find_by_value(uid, token, t_asset)
            if job and job.get("status") != t_delete:
                file_filepath = job.get("path")
                file_name = os.path.basename(file_filepath)
                params = job.get("params", dict())
                inline = params.get("inline", False)
                self.__logger.info(
                    {
                        "id": uid,
                        "type": t_asset,
                        "status": t_start,
                        "filepath": file_filepath,
                        "token": token,
                    }
                )
                try:
                    await self.__jobs.update(
                        uid, file_filepath, t_asset, token, params, t_health, 1
                    )
                    res = await response.file(
                        location=file_filepath,
                        filename=(file_name if not inline else None),
                    )
                    self.__logger.info(
                        {
                            "id": uid,
                            "type": t_asset,
                            "status": t_send,
                            "filepath": file_filepath,
                            "token": token,
                        }
                    )
                    return res
                except Exception as e:
                    self.__logger.error(
                        {
                            "id": uid,
                            "type": t_asset,
                            "status": t_error,
                            "filepath": file_filepath,
                            "token": token,
                            "error": str(e),
                        }
                    )
                    await self.__jobs.update(
                        uid, file_filepath, t_asset, token, params, t_error, 1
                    )
                    raise ServerError({"id": uid, "error": e}, status_code=404)
            self.__logger.error(
                {
                    "id": uid,
                    "type": t_asset,
                    "status": t_error,
                    "error": "Cannot find your token",
                    "token": token,
                }
            )
            raise ServerError(
                {"id": uid, "error": "Cannot find your token", "token": token},
                status_code=404,
            )



from notebook.services.contents.filemanager import FileContentsManager as FCM
from naas.onboarding import download_file, wp_set_for_open_filebrowser
from sanic.response import redirect, json
from sanic.views import HTTPMethodView
from naas.runner.env_var import n_env
from naas.types import t_downloader
import traceback
import uuid


class DownloaderController(HTTPMethodView):
    __logger = None

    def __init__(self, logger, *args, **kwargs):
        super(DownloaderController, self).__init__(*args, **kwargs)
        self.__logger = logger

    async def get(self, request):
        uid = str(uuid.uuid4())
        url = str(request.args.get("url", ""))
        mode_api = request.args.get("api", None)
        create = str(request.args.get("create", ""))
        redirect_to = None
        if create and create != "":
            try:
                notebook_fname = f"{create}.ipynb"
                FCM().new(path=notebook_fname)
                wp_set_for_open_filebrowser(notebook_fname)
                redirect_to = f"{n_env.user_url}/lab"
            except Exception as e:
                tb = traceback.format_exc()
                self.__logger.error(
                    {"id": uid, "type": t_downloader, "status": "send", "filepath": url}
                )
                return json({"status": e, "tb": str(tb)})
        else:
            try:
                file_name = download_file(url)
                wp_set_for_open_filebrowser(file_name)
                self.__logger.info(
                    {"id": uid, "type": t_downloader, "status": "send", "filepath": url}
                )
                redirect_to = f"{n_env.user_url}/lab"
            except Exception as e:
                tb = traceback.format_exc()
                self.__logger.error(
                    {"id": uid, "type": t_downloader, "status": "send", "filepath": url}
                )
                return json({"status": e, "tb": str(tb)})
        if mode_api is None:
            return redirect(redirect_to)
        else:
            return json({"status": "ok"})



# Copyright (c) Naas Team.
# Distributed under the terms of the GNU AGPL License.
from .types import t_tz, t_size, error_busy, error_reject, copy_button
from IPython.core.display import display, Javascript, HTML
from .runner.notifications import Notifications
from .runner.callback import Callback
from .dependency import Dependency
from .runner.env_var import n_env
from .runner.proxy import Domain
from .scheduler import Scheduler
import ipywidgets as widgets
from .assets import Assets
from .secret import Secret
from .runner import Runner
from .api import Api
import requests
import os

__version__ = "1.3.12"
__github_repo = "jupyter-naas/naas"
__doc_url = "https://naas.gitbook.io/naas/"
__canny_js = '<script>!function(w,d,i,s){function l(){if(!d.getElementById(i)){var f=d.getElementsByTagName(s)[0],e=d.createElement(s);e.type="text/javascript",e.async=!0,e.src="https://canny.io/sdk.js",f.parentNode.insertBefore(e,f)}}if("function"!=typeof w.Canny){var c=function(){c.q.push(arguments)};c.q=[],w.Canny=c,"complete"===d.readyState?l():w.attachEvent?w.attachEvent("onload",l):w.addEventListener("load",l,!1)}}(window,document,"canny-jssdk","script");</script>'  # noqa: E501
__crisp = '<script type="text/javascript">window.$crisp=[];window.CRISP_WEBSITE_ID="a64b999e-e44c-44ee-928f-5cd0233f9586";(function(){d=document;s=d.createElement("script");s.src="https://client.crisp.chat/l.js";s.async=1;d.getElementsByTagName("head")[0].appendChild(s);})();</script>'  # noqa: E501
__location__ = os.getcwd()

scheduler = Scheduler()
secret = Secret()
runner = Runner()
api = Api(True)
webhook = Api()
assets = Assets()
dependency = Dependency()
notifications = Notifications()
callback = Callback()
Domain = Domain()


def version():
    print(__version__)


n_env.version = __version__


def get_last_version():
    url = f"https://api.github.com/repos/{__github_repo}/tags"
    response = requests.get(url, headers={"Accept": "application/vnd.github.v3+json"})
    return response.json()[0]["name"]


def get_size():
    response = requests.get(f"{n_env.api}/{t_size}")
    return response.json()


def changelog():
    data = __canny_js
    data += """<button class="lm-Widget p-Widget jupyter-widgets jupyter-button widget-button mod-primary" data-canny-changelog>
        View Changelog
    </button>"""
    data += "<script> Canny('initChangelog', {appID: '5f81748112b5d73b2faf4b15', position: 'bottom', align: 'left'});</script>"
    display(HTML(data))


def open_help():
    data = __crisp
    data += (
        f'<script>$crisp.push(["set", "user:email", ["{str(n_env.user)}"]])</script>'
    )
    data += f'<script>$crisp.push(["set", "session:data", [[["naas_version", "{str(n_env.version)}"]]]])</script>'
    data += '<script>$crisp.push(["do", "chat:open"])</script>'
    display(HTML(data))


def close_help():
    data = '<script>$crisp.push(["do", "chat:hide"])</script>'
    display(HTML(data))


def bug_report():
    email = n_env.user
    name = email.split(".")[0]
    name = email.split("@")[0]
    board_id = "6a83d5c5-2165-2608-082d-49959c7f030c"

    data = __canny_js
    data += "<div data-canny />"
    data += """
    <script>
        Canny('identify', {
            appID: '5f81748112b5d73b2faf4b15',
            user: {
                email: "{EMAIL}",
                name: "{NAME}",
                created: new Date().toISOString()
            },
        });
        Canny('render', {
            boardToken: "{BOARD}",
        });
    </script>
    """

    data = data.replace("{EMAIL}", str(n_env.user))
    data = data.replace("{BOARD}", board_id)
    data = data.replace("{NAME}", name)
    display(HTML(data))


def feature_request():
    email = str(n_env.user)
    name = email.split(".")[0]
    name = email.split("@")[0]
    board_id = "e3e3e0c3-7520-47f5-56f5-39182fb70480"

    data = __canny_js
    data += "<div data-canny />"
    data += """
    <script>
        Canny('identify', {
            appID: '5f81748112b5d73b2faf4b15',
            user: {
                email: "{EMAIL}",
                name: "{NAME}",
                created: new Date().toISOString()
            },
        });
        Canny('render', {
            boardToken: "{BOARD}",
        });
    </script>
    """

    data = data.replace("{EMAIL}", str(n_env.user))
    data = data.replace("{BOARD}", board_id)
    data = data.replace("{NAME}", name)
    display(HTML(data))


def doc():
    button = widgets.Button(description="Open Doc", button_style="primary")
    output = widgets.Output()

    def on_button_clicked(b):
        with output:
            display(Javascript('window.open("{url}");'.format(url=__doc_url)))

    button.on_click(on_button_clicked)
    display(button, output)


def up_to_date():
    return get_last_version() == version()


def update():
    username = n_env.user
    api_url = f"{n_env.hub_api}/hub/api"
    r = requests.delete(
        f"{api_url}/users/{username}/server",
        headers={
            "Authorization": f"token {n_env.token}",
        },
    )
    r.raise_for_status()
    return r


def auto_update():
    if not up_to_date():
        update()
    else:
        print("You are aready up to date")


def get_download_url(url):
    dl_url = f"{n_env.any_user_url}/naas/downloader?url={url}"
    print("❤️ Copy this url and spread it to the world\n")
    copy_button(dl_url)
    return dl_url


def is_production():
    return api.manager.is_production()


def remote_connect(user, token):
    n_env.token = token
    n_env.user = user


def get_remote_timezone():
    try:
        r = requests.get(f"{n_env.api}/{t_tz}")
        r.raise_for_status()
        res = r.json()
        print(f"🕣 Your Production Timezone is {res.get('tz')}\n")
        return res
    except requests.exceptions.ConnectionError as err:
        print(error_busy, err)
        raise
    except requests.exceptions.HTTPError as err:
        print(error_reject, err)
        raise


def set_remote_timezone(timezone):
    n_env.tz = timezone
    try:
        r = requests.post(f"{n_env.api}/{t_tz}", json={"tz": timezone})
        r.raise_for_status()
        res = r.json()
        print(f"🕣 Your Production Timezone is {res.get('tz')}\n")
        return res
    except requests.exceptions.ConnectionError as err:
        print(error_busy, err)
        raise
    except requests.exceptions.HTTPError as err:
        print(error_reject, err)
        raise



from naas.runner.env_var import n_env
import urllib.parse
import requests
import os

__jup_def_set_workspace = "/etc/naas/set_workspace.json"
__jup_def_set_workspace_browser = "/etc/naas/set_workspace_browser.json"
__jup_load_workspace = "jupyter lab workspaces import "
__github_repo = "jupyter-naas/starters"
__github_brach = "main"
__github_api_url = "https://api.github.com/repos/{REPO}/git/trees/{BRANCH}?recursive=1"
__github_base_url = "https://github.com/{REPO}/blob/{BRANCH}/"


def download_file(url):
    raw_target = url
    file_name = raw_target.split("/")[-1]
    file_name = urllib.parse.unquote(file_name)
    file_name = f"dl_{file_name}"
    if "github.com" in raw_target:
        raw_target = raw_target.replace(
            "https://github.com/", "https://raw.githubusercontent.com/"
        )
        raw_target = raw_target.replace("/blob/", "/")
    content = b"ERROR"
    if "://" not in raw_target:
        try:
            cur_path = os.path.join(n_env.server_root, raw_target)
            ff = open(cur_path, "rb")
            content = ff.read()
            ff.close()
        except Exception as e:
            print("Cannot open local file", e)
    else:
        r = requests.get(raw_target)
        content = r.content

    with open(file_name, "wb") as f:
        f.write(content)
        f.close()
    return file_name


def wp_set_for_open_filebrowser(url):
    try:
        filename = url.split("/")[-1]
        filename = filename.split(".")[0]
        new_wp = os.path.join(n_env.path_naas_folder, f"{filename}_workspace.json")
        with open(__jup_def_set_workspace_browser, "r") as fh:
            content_wp = fh.read()
            new_content_wp = content_wp.replace("{NB_NAME}", filename)
            with open(new_wp, "w+") as f:
                f.write(new_content_wp)
        os.system(f"{__jup_load_workspace} {new_wp}")
    except Exception as e:
        print("Cannot config jupyter workspace", e)


def __wp_set_for_open(url):
    try:
        filename_full = url.split("/")[-1]
        filename_num = filename_full.split(".")[0]
        filename = filename_num.split("__")[1]
        new_wp = os.path.join(n_env.path_naas_folder, f"{filename}_workspace.json")
        if not os.path.exists(new_wp):
            old_filename = download_file(url)
            os.system(f"mv {old_filename} {filename}.ipynb")
            with open(__jup_def_set_workspace, "r") as fh:
                content_wp = fh.read()
                new_content_wp = content_wp.replace("{NB_NAME}", filename)
                with open(new_wp, "w+") as f:
                    f.write(new_content_wp)
            os.system(f"{__jup_load_workspace} {new_wp}")
    except Exception as e:
        print("Cannot config jupyter workspace", e)


def __get_onboarding_list():
    url = __github_api_url.replace("{REPO}", __github_repo).replace(
        "{BRANCH}", __github_brach
    )
    url_list = []
    try:
        r = requests.get(url)
        data = r.json()
        for ff in data.get("tree"):
            path = ff.get("path")
            if not path.startswith(".") and path.endswith(".ipynb"):
                base = __github_base_url.replace("{REPO}", __github_repo).replace(
                    "{BRANCH}", __github_brach
                )
                good_url = f"{base}{path}"
                url_list.append(good_url)
    except Exception as e:
        print("__get_onboarding_list", e)
    return url_list


def init_onborading():
    #  jupyter lab workspaces import file_name.json
    try:
        if os.path.exists(n_env.custom_path):
            print("In Naas Docker machine")
            file_list = __get_onboarding_list()
            for url in file_list:
                try:
                    __wp_set_for_open(url)
                except Exception as e:
                    print("error for", url, e)
    except Exception as e:
        print("Cannot config jupyter", e)
